#
# Image Steganography Example
#

# Takes two pictures as inputs and returns a new pictures that has the second
# pictures "hidden" inside the first. The hiding is performed by:
#
#   1. Converting the second pictures (tohide) to grayscale by averaging the
#      red, green, and blue components.
#
#   2. Quantizing the 256 gray levels into 8 "buckets" (0-31, 32-63, ...,
#      224-255) so that they can be represented using 3 bits instead of 8.
#
#   3. Clear out the 3 least significant bits of the green channel from the
#      main picture and adding in the corresponding value from the previous
#      step.
#
# This function assumes that both pictures have the same dimensions. A
# good exercise for students would be to have them make it work for images
# of different sizes (perhaps by taking the maximum of each dimension, forcing
# them to confront bounds checking).
#
def hidePicture(main, tohide):
  w = getWidth(main)
  h = getHeight(main)

  newpicture = makeEmptyPicture(w, h)
  
  for y in range(h):
    for x in range(w):
      mainpixel = getPixel(main, x, y)
      hidepixel = getPixel(tohide, x, y)

      # Step 1: Convert to grayscale
      hidegray = (getRed(hidepixel) + getGreen(hidepixel) + getBlue(hidepixel)) / 3
      
      # Step 2: Quantize
      hidegray = hidegray / 32

      # Step 3: Clear out low bits of green and replace with gray value
      maingreen = getGreen(mainpixel)
      newgreen = int(maingreen / 8) * 8 + hidegray
      
      # Finally, update the pixel in the result image
      newpixel = getPixel(newpicture, x, y)
      setRed(newpixel, getRed(mainpixel))
      setGreen(newpixel, newgreen)
      setBlue(newpixel, getBlue(mainpixel))

  return newpicture


# This function "undoes" the hidePicture function, at least to the best of its
# ability. It takes as its input a picture that has another picture hidden
# inside it and returns the hidden picture.
#
# Since hidePicture converts the picture to hide to grayscale and then reduces
# the number of gray levels from 256 to 8, this provides a good opportunity to
# discuss data loss with students. When they run this function, they'll see
# that the extracted picture is very low quality but still somewhat
# discernable.
#
def findHiddenPicture(pic):
  w = getWidth(pic)
  h = getHeight(pic)
  
  newimage = makeEmptyPicture(w, h)
  
  for y in range(h):
    for x in range(w):
      p = getPixel(pic, x, y)
      
      bri = getGreen(p) % 8 * 32
      
      newp = getPixel(newimage, x, y)
      setRed(newp, bri)
      setGreen(newp, bri)
      setBlue(newp, bri)
      
  return newimage


# Load a couple of pictures that we'll use for the example.
rick = makePicture('/Users/allevato/mediacomp/rick.png')
meadow = makePicture('/Users/allevato/mediacomp/meadow.png')

# Hide Rick in the meadow and show the result. It's impossible to see Rick
# in the final picture because the hidden image manifests itself as subtle
# noise.
newimage = hidePicture(meadow, rick)
show(newimage)

# Now reverse the process; take Rick hiding in the meadow and bring him
# back out.
hidden = findHiddenPicture(newimage)
show(hidden)
